﻿using ColossalFramework;
using ColossalFramework.Plugins;
using ColossalFramework.Math;
using ICities;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Xml;
using UnityEngine;

namespace SharpJunctionAngles
{

    public class AIreplacer
    {

        private class InfoAIpair
        {
            public NetInfo info;
            public NetAI replacedAI;
            public InfoAIpair(NetInfo info, NetAI replacedAI) { this.info = info; this.replacedAI = replacedAI; }
        }

        System.Type oldAItype = null;
        System.Type newAItype = null;
        List<InfoAIpair> replacedInstances = new List<InfoAIpair>();


        public AIreplacer(System.Type oldAItype, System.Type newAItype)
        {
            this.oldAItype = oldAItype;
            this.newAItype = newAItype;
        }


        public ElevationThresholdChanger[] replaceAI(NetInfo[] netInfos)
        {

            GameObject parent = GameObject.Find(SharpJunctionAnglesLoader.PARENT_GAME_OBJECT_NAME);
            if (!parent)
                parent = new GameObject(SharpJunctionAnglesLoader.PARENT_GAME_OBJECT_NAME);

            for (int i = 0; i < netInfos.Length; ++i)
                if (netInfos[i].m_netAI.GetType() == oldAItype)
                {
                    replacedInstances.Add(new InfoAIpair(netInfos[i], netInfos[i].m_netAI));
                    netInfos[i].m_netAI = parent.AddComponent(newAItype) as NetAI;
                    Util.copyCommonFieldsByName(replacedInstances.Last().replacedAI, netInfos[i].m_netAI);
                }

            ElevationThresholdChanger[] result = new ElevationThresholdChanger[replacedInstances.Count];
            for (int i = 0; i < replacedInstances.Count; ++i)
                result[i] = replacedInstances[i].info.m_netAI as ElevationThresholdChanger;
            return result;

        }


        public void revertReplacement()
        {
            for (int i = 0; i < replacedInstances.Count(); ++i)
                replacedInstances[i].info.m_netAI = replacedInstances[i].replacedAI;
            replacedInstances.Clear();
        }


    }


    public interface ElevationThresholdChanger
    {
        void setElevationThreshold(float threshold);
        float getElevationTrheshold();
    }



    public class RoadAIext : RoadAI, ElevationThresholdChanger
    //public class NetAIext<T> : T, ElevationThresholdChanger //TODO
    {
        public static AIreplacer replacer = new AIreplacer(typeof(RoadAI), typeof(RoadAIext));

        //public override void GetElevationLimits(out int min, out int max)
        //{
        //    min = -100;// ((this.m_tunnelInfo == null || (Singleton<ToolManager>.instance.m_properties.m_mode & ItemClass.Availability.GameAndAsset) == ItemClass.Availability.None) ? 0 : SharpJunctionAngleBehaviour.MIN_HEIGHT);
        //    max = +100;// ((this.m_elevatedInfo == null && this.m_bridgeInfo == null) ? 0 : SharpJunctionAngleBehaviour.MAX_HEIGHT);
        //}

        private float elevationThreshold = 8.0f;

        void ElevationThresholdChanger.setElevationThreshold(float threshold)
        {
            elevationThreshold = threshold;
        }

        float ElevationThresholdChanger.getElevationTrheshold()
        {
            return elevationThreshold;
        }

        //public override void GetTerrainModifyRange(out float start, out float end)
        //{
        //    start = 0f;
        //    end = 10f;
        //}

        //private static bool ForceValidDirection(NetInfo info, Vector3 position, ref Vector3 direction, ushort segmentID, ref NetSegment segment)
        //{
        //    return true;
        //}

        //public override ToolBase.ToolErrors CheckBuildPosition(bool test, bool visualize, bool overlay, bool autofix, ref NetTool.ControlPoint startPoint, ref NetTool.ControlPoint middlePoint, ref NetTool.ControlPoint endPoint, out BuildingInfo ownerBuilding, out Vector3 ownerPosition, out Vector3 ownerDirection, out int productionRate)
        //{
        //    NetTool.ControlPoint dummyStart = new NetTool.ControlPoint();
        //    NetTool.ControlPoint dummyMiddle = new NetTool.ControlPoint();
        //    NetTool.ControlPoint dummyEnd = new NetTool.ControlPoint();
        //    base.CheckBuildPosition(false, visualize, overlay, false, ref dummyStart, ref dummyMiddle, ref dummyEnd, out ownerBuilding, out ownerPosition, out ownerDirection, out productionRate);
        //    return ToolBase.ToolErrors.None;
        //}




        public override NetInfo GetInfo(float minElevation, float maxElevation, float length, bool incoming, bool outgoing, bool curved, bool enableDouble, ref ToolBase.ToolErrors errors)
        {
            //TODO: generalize this to arbitrary NetAI descendants by checking field presence per reflection
            //NetAI thisNetAI = this as NetAI;//this;

            if (incoming || outgoing)
            {
                int num;
                int num2;
                Singleton<BuildingManager>.instance.CalculateOutsideConnectionCount(this.m_info.m_class.m_service, this.m_info.m_class.m_subService, out num, out num2);
                if ((incoming && num >= 4) || (outgoing && num2 >= 4))
                {
                    errors |= ToolBase.ToolErrors.TooManyConnections;
                }
            }
            if (maxElevation > 255f)
            {
                errors |= ToolBase.ToolErrors.HeightTooHigh;
            }
            if ((Singleton<ToolManager>.instance.m_properties.m_mode & ItemClass.Availability.GameAndAsset) != ItemClass.Availability.None)
            {
                if (this.m_tunnelInfo != null && maxElevation < -8)
                {
                    return this.m_tunnelInfo;
                }
                if (this.m_slopeInfo != null && minElevation < -8f)
                {
                    return this.m_slopeInfo;
                }
            }
            if (this.m_bridgeInfo != null && maxElevation > 25f && length > 45f && !curved && (enableDouble || !this.m_bridgeInfo.m_netAI.RequireDoubleSegments()))
            {
                return this.m_bridgeInfo;
            }
            if (this.m_elevatedInfo != null && maxElevation > elevationThreshold)
            {
                return this.m_elevatedInfo;
            }
            if (maxElevation > elevationThreshold)
            {
                errors |= ToolBase.ToolErrors.HeightTooHigh;
            }
            return this.m_info;
        }

    }


    public class TrainTrackAIext : TrainTrackAI, ElevationThresholdChanger
    {
        public static AIreplacer replacer = new AIreplacer(typeof(TrainTrackAI), typeof(TrainTrackAIext));

        private float elevationThreshold = 8.0f;

        void ElevationThresholdChanger.setElevationThreshold(float threshold)
        {
            elevationThreshold = threshold;
        }

        float ElevationThresholdChanger.getElevationTrheshold()
        {
            return elevationThreshold;
        }

        public override NetInfo GetInfo(float minElevation, float maxElevation, float length, bool incoming, bool outgoing, bool curved, bool enableDouble, ref ToolBase.ToolErrors errors)
        {
            //TODO: generalize this to arbitrary NetAI descendants by checking field presence per reflection
            //NetAI thisNetAI = this as NetAI;//this;

            if (incoming || outgoing)
            {
                int num;
                int num2;
                Singleton<BuildingManager>.instance.CalculateOutsideConnectionCount(this.m_info.m_class.m_service, this.m_info.m_class.m_subService, out num, out num2);
                if ((incoming && num >= 4) || (outgoing && num2 >= 4))
                {
                    errors |= ToolBase.ToolErrors.TooManyConnections;
                }
            }
            if (maxElevation > 255f)
            {
                errors |= ToolBase.ToolErrors.HeightTooHigh;
            }
            if ((Singleton<ToolManager>.instance.m_properties.m_mode & ItemClass.Availability.GameAndAsset) != ItemClass.Availability.None)
            {
                if (this.m_tunnelInfo != null && maxElevation < -8)
                {
                    return this.m_tunnelInfo;
                }
                if (this.m_slopeInfo != null && minElevation < -8f)
                {
                    return this.m_slopeInfo;
                }
            }
            if (this.m_bridgeInfo != null && maxElevation > 25f && length > 45f && !curved && (enableDouble || !this.m_bridgeInfo.m_netAI.RequireDoubleSegments()))
            {
                return this.m_bridgeInfo;
            }
            if (this.m_elevatedInfo != null && maxElevation > elevationThreshold)
            {
                return this.m_elevatedInfo;
            }
            if (maxElevation > elevationThreshold)
            {
                errors |= ToolBase.ToolErrors.HeightTooHigh;
            }
            return this.m_info;
        }

    }


    public class PedestrianPathAIext : PedestrianPathAI, ElevationThresholdChanger
    {
        public static AIreplacer replacer = new AIreplacer(typeof(PedestrianPathAI), typeof(PedestrianPathAIext));

        private float elevationThreshold = 8.0f;

        void ElevationThresholdChanger.setElevationThreshold(float threshold)
        {
            elevationThreshold = threshold;
        }

        float ElevationThresholdChanger.getElevationTrheshold()
        {
            return elevationThreshold;
        }

        public override NetInfo GetInfo(float minElevation, float maxElevation, float length, bool incoming, bool outgoing, bool curved, bool enableDouble, ref ToolBase.ToolErrors errors)
        {
            //TODO: generalize this to arbitrary NetAI descendants by checking field presence per reflection
            //NetAI thisNetAI = this as NetAI;//this;

            if (incoming || outgoing)
            {
                int num;
                int num2;
                Singleton<BuildingManager>.instance.CalculateOutsideConnectionCount(this.m_info.m_class.m_service, this.m_info.m_class.m_subService, out num, out num2);
                if ((incoming && num >= 4) || (outgoing && num2 >= 4))
                {
                    errors |= ToolBase.ToolErrors.TooManyConnections;
                }
            }
            if (maxElevation > 255f)
            {
                errors |= ToolBase.ToolErrors.HeightTooHigh;
            }
            if ((Singleton<ToolManager>.instance.m_properties.m_mode & ItemClass.Availability.GameAndAsset) != ItemClass.Availability.None)
            {
                if (this.m_tunnelInfo != null && maxElevation < -8)
                {
                    return this.m_tunnelInfo;
                }
                if (this.m_slopeInfo != null && minElevation < -8f)
                {
                    return this.m_slopeInfo;
                }
            }
            if (this.m_bridgeInfo != null && maxElevation > 25f && length > 45f && !curved && (enableDouble || !this.m_bridgeInfo.m_netAI.RequireDoubleSegments()))
            {
                return this.m_bridgeInfo;
            }
            if (this.m_elevatedInfo != null && maxElevation > elevationThreshold)
            {
                return this.m_elevatedInfo;
            }
            if (maxElevation > elevationThreshold)
            {
                errors |= ToolBase.ToolErrors.HeightTooHigh;
            }
            return this.m_info;
        }

    }


}