﻿using ColossalFramework;
using ColossalFramework.Plugins;
using ColossalFramework.Math;
using ICities;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Xml;
using UnityEngine;

namespace SharpJunctionAngles//some sources say unity does not support namespaces for MonoBehaviours
{

    public class SharpJunctionAnglesLoader : LoadingExtensionBase
    {
        public const string PARENT_GAME_OBJECT_NAME = "SharpJunctionAngles";

        GameObject anarchyUpdate;
        SharpJunctionAngleBehaviour anarchy;
        
        public override void OnLevelLoaded(LoadMode mode)
        {
            //base.OnLevelLoaded(mode);
            anarchyUpdate = new GameObject(PARENT_GAME_OBJECT_NAME);
            anarchy = anarchyUpdate.AddComponent<SharpJunctionAngleBehaviour>();
            //anarchy.EnableHook();

            //DebugOutputPanel.AddMessage(PluginManager.MessageType.Message, "SharpJunctionAngles loaded: press L to toggle anarchy and K to toggle junction angles.");
            
        }

        public override void OnLevelUnloading()
        {
            anarchy.enableSegmentSnapping();
            anarchy.disableAnarchy();
            anarchy = null;
            GameObject.Destroy(anarchyUpdate);
            base.OnLevelUnloading();
        }

        //just for debugging while updating the mod at runtime
        //  -> doesn't work at all, the games keeps instantiating an outdated version of SharpJunctionAngleBehaviour and AddComponent returns null...
        //  (also no proper deinit (OnReleased() doesn't get called) is performed when replacing the dll, which might cause the problem)
        /*
        public override void OnCreated(ILoading loading)
        {
            SharpJunctionAngleBehaviour[] anarchies = Resources.FindObjectsOfTypeAll(typeof(SharpJunctionAngleBehaviour)) as SharpJunctionAngleBehaviour[];
            DebugOutputPanel.AddMessage(PluginManager.MessageType.Message, "anarchies.length(): " + anarchies.Length);
            for (int i = 0; i < anarchies.Length; ++i)
                GameObject.Destroy(anarchies[i]);
            GameObject[] gameObjects = Resources.FindObjectsOfTypeAll(typeof(GameObject)) as GameObject[];
            for (int i = 0; i < gameObjects.Length; ++i)
                if (gameObjects[i].name == "BendedSegmentDisabler")
                    GameObject.Destroy(gameObjects[i]);
                       
            anarchyUpdate = new GameObject("BendedSegmentDisabler");
            //anarchy = anarchyUpdate.AddComponent<Anarchy>();
            anarchy = anarchyUpdate.AddComponent<SharpJunctionAngleBehaviour>();
            DebugOutputPanel.AddMessage(PluginManager.MessageType.Message, "anarchy == null: " + (anarchy == null));
            
            if(anarchy)
                anarchy.EnableHook();

            base.OnCreated(loading);

        }
        */
    }  
   


    public class SharpJunctionAngleBehaviour : MonoBehaviour
    {

        public const int MAX_HEIGHT =  1000;//5
        public const int MIN_HEIGHT = -1000;//-3
          
                        
        //keybindings and default values
        KeyBinding anarchyBinding           = new KeyBinding(KeyCode.LeftControl, KeyCode.A, KeyCode.None);
        KeyBinding bendingBinding           = new KeyBinding(KeyCode.LeftControl, KeyCode.B, KeyCode.None);
        KeyBinding snappingBinding          = new KeyBinding(KeyCode.LeftControl, KeyCode.S, KeyCode.None);
        KeyBinding elevationBinding         = new KeyBinding(KeyCode.LeftControl, KeyCode.R, KeyCode.None);
        KeyBinding groundInteractionBinding = new KeyBinding(KeyCode.LeftControl, KeyCode.G, KeyCode.None);
        

        private bool anarchyEnabled = false;
        private Dictionary<MethodInfo, Detours.RedirectCallsState> redirects = new Dictionary<MethodInfo, Detours.RedirectCallsState>();
        private string[] hookedToolTypeNameFilters =  {"Net","net"};//type name filter with default values
        private System.Type[] detouredToolTypes;
        private System.Type[] detouredNetAItypes;
        //private MethodInfo createNodeMethodInfo;
        //private RedirectCallsState createNodeRedirectCallState;

        private bool snappingEnabled = true;
        private Dictionary<MethodInfo, Detours.RedirectCallsState> redirectsSnapping = new Dictionary<MethodInfo, Detours.RedirectCallsState>();

        private NetInfo[] netInfos;

        private bool bendingDisabled = false;
        private NetInfo[] bendingNetInfos;

        private bool disableObjectCollisions = true;
        private bool elevationLimitChecksDisabled = true;
        private bool enlargeCatenaryMastDistance = true;
        private bool groundInteractionDisabled = false;
        private bool hideStatusPanel = false;

        private ColossalFramework.UI.UIButton chirperButton;
        private ColossalFramework.UI.UILabel labelAnarchy = null, labelBending = null, labelElevation = null, labelSnapping = null;

        private bool lastNetToolEnabled = true;

        //private float terrainStep = 12;

        static public void Log(string message)
        {
            Debug.Log(message);
            DebugOutputPanel.AddMessage(PluginManager.MessageType.Message, message);
        }

        SharpJunctionAngleBehaviour():
            base()
        {

            Log("SharpJunctionAngleBehaviour(): reading config...");
            readConfig();

            Log("SharpJunctionAngleBehaviour(): searching netinfos...");
            bendingNetInfos = findBendingSegmentNetInfos();

            Log("SharpJunctionAngleBehaviour(): searching concrete ToolBase subclasses...");
            try
            {
                for (int i = 0; i < hookedToolTypeNameFilters.Length; ++i)
                    Log(hookedToolTypeNameFilters[i]);

                detouredToolTypes = Util.findConcreteSubclasses(typeof(ToolBase)).Where(type => Util.containsOneOf(type.Name, hookedToolTypeNameFilters)).ToArray();
            }
            catch (Exception e)
            {
                Log("Failed to retrieve ToolBase subclasses! Only trying to use NetTool or NetToolFine.");
                Log(" -> This may be caused by an outdated Fine Road Heights mod.");
                Log(" Failed with exception: " + e.Message);
                detouredToolTypes = new Type[1];
                Type netToolFineType = System.Type.GetType("NetToolFine");
                if(netToolFineType != null)
                    detouredToolTypes[0] = netToolFineType;
                else
                    detouredToolTypes[0] = typeof(NetTool);
            }

            Log("SharpJunctionAngleBehaviour(): searching concrete NetAI subclasses...");
            detouredNetAItypes = Util.findConcreteSubclasses(typeof(NetAI));
            
            //terrainStep = getTerrainStepFromNetTool();

            Log("SharpJunctionAngleBehaviour(): initUI()...");
            initUI();
            setUIvisibility(!hideStatusPanel);
                       

            if (enlargeCatenaryMastDistance)
            {
                Log("SharpJunctionAngleBehaviour(): enlargeCateneryMastDistance()");
                enlargeCateneryMastDistance();
            }

            //RoadAIext.replacer.replaceRoadAI(netInfos);
        }
        

        private NetInfo[] findBendingSegmentNetInfos()
        { 
            netInfos = Resources.FindObjectsOfTypeAll(typeof(NetInfo)) as NetInfo[];
            List<NetInfo> bendingNetInfos = new List<NetInfo>(100);
            for (int i = 0; i < netInfos.Length; ++i)
                if (netInfos[i].m_enableBendingSegments)
                    bendingNetInfos.Add(netInfos[i]);
            DebugOutputPanel.AddMessage(PluginManager.MessageType.Message, "SharpJunctionAngles V0.4: Found " + bendingNetInfos.Count + " bending NetInfos.");
            return bendingNetInfos.ToArray();           
        }
        
        private ColossalFramework.UI.UIButton findChirperButton()
        {
            var chirperPanelZone = GameObject.Find("ChirperPanel/Zone");
            return chirperPanelZone != null ? chirperPanelZone.GetComponent<ColossalFramework.UI.UIButton>() : null;
        }


        private ToolBase getEnabledTool()
        {
            var toolController = GameObject.Find("Tool Controller");
            if (!toolController)
                return null;
            for (int i = 0; i < detouredToolTypes.Length; ++i)
            {
                var tool = toolController.GetComponent(detouredToolTypes[i].Name) as ToolBase;
                if (tool != null && tool.enabled)
                    return tool;
            }
            return null;
        }

        private float getTerrainStepFromNetTool()
        {
            var toolController = GameObject.Find("Tool Controller");
            if(toolController)
                for(int i=0; i<detouredToolTypes.Length; ++i){

                    var tool = toolController.GetComponent(detouredToolTypes[i].Name);
                    var fieldInfo = detouredToolTypes[i].GetField("m_terrainStep", Util.ALL_METHOD_FLAGS);
           
                    if ( tool != null && fieldInfo != null )                    
                        return (float)fieldInfo.GetValue(tool);                    

                }

            return 12;
        }


        private bool isNetToolEnabled()
        {
            return getEnabledTool() != null;
        }
        

        private void enlargeCateneryMastDistance()
        {
            //remove all lane props
            //for (int i = 0; i < m_bendingNetInfos.Length; ++i)
            //    for (int j = 0; j < m_bendingNetInfos[i].m_lanes.Length; ++j)
            //        if (m_bendingNetInfos[i].m_lanes[j].m_laneProps)
            //            m_bendingNetInfos[i].m_lanes[j].m_laneProps.m_props = new NetLaneProps.Prop[0];

            //change repeat distance
            for(int i=0; i<bendingNetInfos.Length; ++i)
                for(int j=0; j<bendingNetInfos[i].m_lanes.Length; ++j)
                    if (bendingNetInfos[i].m_lanes[j].m_laneProps)
                    {
                        for (int k = 0; k < bendingNetInfos[i].m_lanes[j].m_laneProps.m_props.Length; ++k)
                            if (bendingNetInfos[i].m_lanes[j].m_laneProps.m_props[k].m_finalProp
                                && bendingNetInfos[i].m_lanes[j].m_laneProps.m_props[k].m_finalProp.name.IndexOf("Power") > -1)
                            {
                                //m_bendingNetInfos[i].m_lanes[j].m_laneProps.m_props[k].m_repeatDistance = 10.0f;
                                //Debug.Log("mast offset: " + bendingNetInfos[i].m_lanes[j].m_laneProps.m_props[k].m_segmentOffset);
                                if (   bendingNetInfos[i].m_lanes[j].m_laneProps.m_props[k].m_segmentOffset != 1)
                                    bendingNetInfos[i].m_lanes[j].m_laneProps.m_props[k] = null;//.m_segmentOffset = -1;
                                //m_bendingNetInfos[i].m_lanes[j].m_laneProps.m_props[k].m_finalProp.enabled = false;
                            }
                        bendingNetInfos[i].m_lanes[j].m_laneProps.m_props = bendingNetInfos[i].m_lanes[j].m_laneProps.m_props.Where(prop => prop != null).ToArray();
                    }

        }


        private void initUI()
        {
            DebugOutputPanel.AddMessage(PluginManager.MessageType.Message,"initUI");   

            uint labelCount = 0u;
            
            chirperButton = findChirperButton();

            var aspectRatio = (float)UnityEngine.Screen.width / (float)UnityEngine.Screen.height;

            const float LABEL_SPACING = 0.028f;
            float LABEL_POS_X = -aspectRatio+0.025f;//-1.25f;//-1.25f
            float LABEL_POS_Y = -0.675f + Math.Min(0,(int)labelCount-0) * LABEL_SPACING;//-0.62f
            var uiView = ColossalFramework.UI.UIView.GetAView();
                      
            labelAnarchy = (ColossalFramework.UI.UILabel)uiView.AddUIComponent(typeof(ColossalFramework.UI.UILabel));
            labelAnarchy.name = "labelAnarchy";
            labelAnarchy.transformPosition = new Vector3(LABEL_POS_X, LABEL_POS_Y - labelCount * LABEL_SPACING);
            labelAnarchy.tooltip = "Is Road Anarchy enabled? (required for sharp junctions)";
            ++labelCount;

            labelBending = (ColossalFramework.UI.UILabel)uiView.AddUIComponent(typeof(ColossalFramework.UI.UILabel));
            labelBending.name = "labelBending";
            labelBending.transformPosition = new Vector3(LABEL_POS_X, LABEL_POS_Y - labelCount * LABEL_SPACING);
            labelBending.tooltip = "Is auto segment bending enabled? (turn off for sharp junctions)";
            ++labelCount;

            labelSnapping = (ColossalFramework.UI.UILabel)uiView.AddUIComponent(typeof(ColossalFramework.UI.UILabel));
            labelSnapping.name = "labelSnapping";
            labelSnapping.transformPosition = new Vector3(LABEL_POS_X, LABEL_POS_Y - labelCount * LABEL_SPACING);
            labelSnapping.tooltip = "Is segment snapping enabled? (turn off for close parallel segments)";
            ++labelCount;

            //labelElevation = (ColossalFramework.UI.UILabel)uiView.AddUIComponent(typeof(ColossalFramework.UI.UILabel));
            //labelElevation.name = "labelElevation";
            //labelElevation.transformPosition = new Vector3(LABEL_POS_X, LABEL_POS_Y - labelCount * LABEL_SPACING);
            //labelElevation.tooltip = "Are elevation limits in force? (turn on for smoother and lower bridge building)";
            //labelElevation.isVisible = false;
            //++labelCount;

            updateUI();
        }

        private void updateUI()
        {
           if (labelAnarchy == null || labelBending == null || labelSnapping == null)
                return;

            labelAnarchy.text   = anarchyEnabled               ? "anarchy: on"        : "anarchy: off";
            labelBending.text   = bendingDisabled              ? "└bending: off"      : "└bending: on";
            labelSnapping.text  = snappingEnabled              ? "seg. snap: on"      : "seg. snap: off";
            //labelElevation.text = elevationLimitChecksDisabled ? " elevation: free"   : " elevation: limited";

            if (chirperButton)
                chirperButton.transform.localScale = new Vector3(anarchyEnabled ? 2.0f : 1.0f, 1.0f, 1.0f); //chirperButton.focusedColor = new Color32(255, 0, 0, 255);
       
        }
        
        private void setUIvisibility(bool val)
        {
            if ( hideStatusPanel )
                val = false;
          
            labelAnarchy.isVisible = val;
            labelBending.isVisible = val;
            labelSnapping.isVisible = val;
            //labelElevation.isVisible = val;
        }


        private string configFilename()
        {
            //DebugOutputPanel.AddMessage(PluginManager.MessageType.Message, "System.Uri.UnescapeDataString(Assembly.GetAssembly(typeof(SharpJunctionAngleBehaviour)).CodeBase): " + System.Uri.UnescapeDataString(Assembly.GetAssembly(typeof(SharpJunctionAngleBehaviour)).CodeBase));
            //return System.IO.Path.GetDirectoryName(System.Uri.UnescapeDataString(Assembly.GetAssembly(typeof(SharpJunctionAngleBehaviour)).CodeBase)) + "\\SharpJunctionAnglesConfig.xml";
            return "SharpJunctionAnglesConfig.xml";
        }

       private void readConfig()
        {
            try
            {
                XmlDocument doc = new XmlDocument();
                if (System.IO.File.Exists(configFilename()))
                {
                    doc.Load(configFilename());
                    
                    XmlNodeList nodes = doc.SelectNodes("//config/anarchyBinding");
                    if (nodes.Count == 1)
                    {
                        anarchyBinding = new KeyBinding(nodes[0].InnerText);
                        DebugOutputPanel.AddMessage(PluginManager.MessageType.Message, "Read anarchy key binding from config: " + anarchyBinding.toString());
                    }

                    nodes = doc.SelectNodes("//config/bendingBinding");
                    if (nodes.Count == 1)
                    {
                        bendingBinding = new KeyBinding(nodes[0].InnerText);
                        DebugOutputPanel.AddMessage(PluginManager.MessageType.Message, "Read bending key binding from config: " + bendingBinding.toString());
                    }

                    nodes = doc.SelectNodes("//config/snappingBinding");
                    if (nodes.Count == 1)
                    {
                        snappingBinding = new KeyBinding(nodes[0].InnerText);
                        DebugOutputPanel.AddMessage(PluginManager.MessageType.Message, "Read snapping key binding from config: " + bendingBinding.toString());
                    }

                    nodes = doc.SelectNodes("//config/elevationBinding");
                    if (nodes.Count == 1)
                    {
                        elevationBinding = new KeyBinding(nodes[0].InnerText);
                        DebugOutputPanel.AddMessage(PluginManager.MessageType.Message, "Read elevation limits key binding from config: " + bendingBinding.toString());
                    }

                    nodes = doc.SelectNodes("//config/disableObjectCollisions");
                    if (nodes.Count == 1)
                    {
                        try
                        {
                            disableObjectCollisions = bool.Parse(nodes[0].InnerText);
                        }
                        catch { Debug.Log("Could not read disableObjectCollisions switch from config."); };
                    }

                    nodes = doc.SelectNodes("//config/enlargeCaternaryMastDistance");
                    if (nodes.Count == 1)
                    {
                        try
                        {
                            enlargeCatenaryMastDistance = bool.Parse(nodes[0].InnerText);
                        }
                        catch { Debug.Log("Could not read caternary mast distance switch from config."); };
                    }

                    nodes = doc.SelectNodes("//config/hideStatusPanel");
                    if (nodes.Count == 1)
                    {
                        try
                        {
                            hideStatusPanel = bool.Parse(nodes[0].InnerText);
                        }
                        catch { Debug.Log("Could not read hideStatusPanel switch from config."); };
                    }

                    nodes = doc.SelectNodes("//config/hookedToolTypeNameFilters");
                    if (nodes.Count == 1)
                    {
                        char[] delimiters = { ' ', '\t' };
                        hookedToolTypeNameFilters = nodes[0].InnerText.Split(delimiters).Where(filter => !filter.IsNullOrWhiteSpace()).ToArray();
                    }

                    supplementConfig(doc);
                    doc.Save(configFilename());
                }
                else
                    writeConfig();
            }
            catch (Exception e)
            {
                Debug.Log("Could not access config file: \"" + configFilename() + "\" Failed with message \"" + e.Message + "\".");
                Debug.Log("Rewriting default config, previous will be renamed to \"" + configFilename() + "_backup" + "\"");
                try
                {
                    System.IO.File.Move(configFilename(), configFilename() + "_backup");
                    writeConfig();
                }
                catch (Exception e2)
                {
                    Debug.Log("Could not rewrite config! Failed with message \"" + e.Message + "\": Using default values!");
                }
            }
            
        }

        private void writeConfig()
        {
            XmlDocument doc = new XmlDocument();
            
            XmlNode config = doc.AppendChild( doc.CreateElement("config") );
                        
            config.AppendChild(doc.CreateComment("Each key binding can be a combination of up to three keys."));
            config.AppendChild(doc.CreateComment("Format: <binding>keycode1 keycode2 keycode3</binding>"));
            config.AppendChild(doc.CreateComment("See http://docs.unity3d.com/ScriptReference/KeyCode.html for valid keycodes."));
            
            config.AppendChild(doc.CreateElement("anarchyBinding")).InnerText = anarchyBinding.toString();
            config.AppendChild(doc.CreateElement("bendingBinding")).InnerText = bendingBinding.toString();
            config.AppendChild(doc.CreateElement("snappingBinding")).InnerText = snappingBinding.toString();            
            //config.AppendChild(doc.CreateElement("elevationBinding")).InnerText = elevationBinding.toString();

            config.AppendChild(doc.CreateComment("If this is true, anarchy also disables object (buildings, trees, props) collisions."));
            config.AppendChild(doc.CreateElement("disableObjectCollisions")).InnerText = disableObjectCollisions.ToString();

            config.AppendChild(doc.CreateComment("If this is true, masts are only cerated at the middle of segments."));
            config.AppendChild(doc.CreateElement("enlargeCaternaryMastDistance")).InnerText = enlargeCatenaryMastDistance.ToString();

            config.AppendChild(doc.CreateComment("If this is true, the status panel in the lower-left screen corner will be hidden."));
            config.AppendChild(doc.CreateElement("hideStatusPanel")).InnerText = hideStatusPanel.ToString();

            config.AppendChild(doc.CreateComment("Name of the affected tool data types is required to contain one of the strings in hookedToolTypeNameFilters."));
            config.AppendChild(doc.CreateComment("(leave empty if you like the side effects on non-network tools)"));
            XmlNode current = config.AppendChild(doc.CreateElement("hookedToolTypeNameFilters"));
            for (int i = 0; i < hookedToolTypeNameFilters.Length; ++i)
                current.InnerText += hookedToolTypeNameFilters[i] + " ";
                     
            
            doc.Save(configFilename());
        }


        private static void ensureNodeUniquePresence(XmlDocument doc, XmlNode parent, string xpath, string comment, string value)
        {
            var nodes = parent.SelectNodes(xpath);
            
            if (nodes.Count != 1)
            {
                if (nodes.Count < 1)
                {
                    if( !comment.IsNullOrWhiteSpace() )
                        parent.AppendChild(doc.CreateComment(comment));
                    parent.AppendChild(doc.CreateElement(xpath)).InnerText = value;
                }
                else
                {                   
                    for (int i = 1; i < nodes.Count; ++i)
                        nodes[i].RemoveAll();
                }
            }

        }


        private void supplementConfig(XmlDocument doc)
        {
            XmlNodeList configs = doc.SelectNodes("//config");
            XmlNode config = configs.Count == 1 ? configs[0] : doc.AppendChild(doc.CreateElement("config"));

            ensureNodeUniquePresence(doc, config, "anarchyBinding", "", anarchyBinding.toString());
            ensureNodeUniquePresence(doc, config, "bendingBinding", "", bendingBinding.toString());
            ensureNodeUniquePresence(doc, config, "snappingBinding", "", snappingBinding.toString());

            ensureNodeUniquePresence(doc, config, "disableObjectCollisions", "If this is true, anarchy also disables building collisions.", disableObjectCollisions.ToString());
            ensureNodeUniquePresence(doc, config, "enlargeCaternaryMastDistance", "If this is true, masts are only cerated at the middle of segments.", enlargeCatenaryMastDistance.ToString());
            ensureNodeUniquePresence(doc, config, "hideStatusPanel", "If this is true, the status panel in the lower-left screen corner will be hidden.", hideStatusPanel.ToString());
            
            string filters = "";
            for (int i = 0; i < hookedToolTypeNameFilters.Length; ++i)
                filters += hookedToolTypeNameFilters[i] + " ";
            ensureNodeUniquePresence(doc, config, "hookedToolTypeNameFilters", "Name of the affected tool data types is required to contain one of the strings in hookedToolTypeNameFilters. \n (leave empty if you like the side effects on non-network tools)", filters);
            
            //if (doc.SelectNodes("//config/anarchyBinding").Count < 1)
            //    config.AppendChild(doc.CreateElement("anarchyBinding")).InnerText = anarchyBinding.toString();
            //if (doc.SelectNodes("//config/bendingBinding").Count < 1)
            //    config.AppendChild(doc.CreateElement("bendingBinding")).InnerText = bendingBinding.toString();
            //if (doc.SelectNodes("//config/snappingBinding").Count < 1)
            //    config.AppendChild(doc.CreateElement("snappingBinding")).InnerText = snappingBinding.toString();            
            ////if (doc.SelectNodes("//config/elevationBinding").Count < 1)
            ////    config.AppendChild(doc.CreateElement("elevationBinding")).InnerText = elevationBinding.toString();

            //if (doc.SelectNodes("//config/disableObjectCollisions").Count < 1)
            //{
            //    config.AppendChild(doc.CreateComment("If this is true, anarchy also disables object (buildings, trees, props) collisions."));
            //    config.AppendChild(doc.CreateElement("disableObjectCollisions")).InnerText = disableObjectCollisions.ToString();
            //}               

            //if (doc.SelectNodes("//config/enlargeCaternaryMastDistance").Count < 1)
            //{
            //    config.AppendChild(doc.CreateComment("If this is true, masts are only cerated at the middle of segments."));
            //    config.AppendChild(doc.CreateElement("enlargeCaternaryMastDistance")).InnerText = enlargeCatenaryMastDistance.ToString();
            //}

            //if (doc.SelectNodes("//config/hookedToolTypeNameFilters").Count < 1)
            //{
            //    config.AppendChild(doc.CreateComment("Name of the affected tool data types is required to contain one of the strings in hookedToolTypeNameFilters."));
            //    config.AppendChild(doc.CreateComment("(leave empty if you like the side effects on non-network tools)"));
            //    XmlNode current = config.AppendChild(doc.CreateElement("hookedToolTypeNameFilters"));
            //    for (int i = 0; i < hookedToolTypeNameFilters.Length; ++i)
            //        current.InnerText += hookedToolTypeNameFilters[i] + " ";
            //}
            
           
        }
       

        public void Update()
        {
                                
            //toggle anarchy
            if( anarchyBinding.isPressed() )
            {
                if (anarchyEnabled)
                    disableAnarchy();
                else{
                    //terrainStep = getTerrainStepFromNetTool();
                    enableAnarchy();
                }
                
                updateUI();
            }

            //toggle Sharp junction angles / auto segment bending
            if( bendingBinding.isPressed() )
            {
                if (bendingDisabled)
                    enableBending();
                else
                    disableBending();

                updateUI();
            }

            //toggle segment snapping
            if (snappingBinding.isPressed())
            {
                if (snappingEnabled)
                    disableSegmentSnapping();
                else
                    enableSegmentSnapping();

                updateUI();
            }



            //toggle anarchy elevation limit checks
            //if (elevationBinding.isPressed())
            //{
            //    if (hookEnabled)
            //    {
            //        DisableHook();
            //        EnableHook();
            //    }
            //    elevationLimitChecksDisabled = !elevationLimitChecksDisabled;
            //    updateUI();
            //}

            //toogle ground interaction 
            //if (groundInteractionBinding.isPressed())
            //{
            //    if (groundInteractionDisabled)
            //        EnableGroundInteraction();
            //    else
            //        DisableGroundInteraction();

            //    updateUI();
            //}
             

            bool netToolEnabled = isNetToolEnabled();
            if (netToolEnabled != lastNetToolEnabled)
                setUIvisibility(isNetToolEnabled());
            lastNetToolEnabled = netToolEnabled;
            
        }
        
       
        

        public void enableAnarchy()
        {
            Log("Enabling Anarchy...");

            if (anarchyEnabled)
                return;
                       
            System.Type targetType = typeof(SharpJunctionAngleBehaviour);
                                    
            for (int i = 0; i < detouredToolTypes.Length; ++i)
            {
                                
                Log("Hooking \"" + detouredToolTypes[i].Name + "\"...");
            
                System.Type toolType = detouredToolTypes[i];
                
                Util.tryMethodRedirection(toolType, "CanCreateSegment", targetType, "CanCreateSegment7", ref redirects, 7);
                Util.tryMethodRedirection(toolType, "CanCreateSegment", targetType, "CanCreateSegment12", ref redirects, 12);
                Util.tryMethodRedirection(toolType, "CanCreateSegment", targetType, "CanCreateSegment13", ref redirects, 13);
                Util.tryMethodRedirection(toolType, "CanAddSegment", targetType, "CanAddSegment", ref redirects);

                Util.tryMethodRedirection(toolType, "CanAddNode", targetType, "CanAddNode4", ref redirects, 4);
                Util.tryMethodRedirection(toolType, "CanAddNode", targetType, "CanAddNode5", ref redirects, 5);

                Util.tryMethodRedirection(toolType, "TestNodeBuilding", targetType, "TestNodeBuilding", ref redirects);
                Util.tryMethodRedirection(toolType, "CheckNodeHeights", targetType, "CheckNodeHeights", ref redirects);
                Util.tryMethodRedirection(toolType, "CheckStartAndEnd", targetType, "CheckStartAndEnd", ref redirects);

                Util.tryMethodRedirection(toolType, "CheckCollidingSegments", targetType, "CheckCollidingSegments", ref redirects);
                Util.tryMethodRedirection(toolType, "CheckCollidingBuildings", targetType, "CheckCollidingBuildings", ref redirects);
                Util.tryMethodRedirection(toolType, "CheckSpace", targetType, "CheckSpace", ref redirects);

                if(toolType.Name == "NetToolFine" )
                    Util.tryMethodRedirection(toolType, "GetAdjustedElevationLimits", targetType, "GetAdjustedElevationLimits", ref redirects);
                

                //Util.tryMethodRedirection(toolType, "Snap", targetType, "Snap", ref redirects);
                //Util.tryMethodRedirection(toolType, "SnapDirection", targetType, "SnapDirection", ref redirects);
                //Util.tryMethodRedirection(toolType, "MoveEndNode", targetType, "MoveEndNode", ref redirects);
                
                //Since some patch, this causes an InvalidOperationException (changed signature or method removed?):
                //tryMethodRedirection(toolType, "CheckZoning", targetType, "CheckZoning", ref redirects, 1);

                //proved unnecessary
                //tryMethodRedirection(toolType, "CreateNode", targetType, "CreateNode", ref redirects, 18);

                //elevtion limits (tool part), conflicts with Fine Road Heights (actually seems unnecessary anyways)
                //tryMethodRedirection(toolType, "GetElevation", targetType, "GetElevation", ref redirects);

                Log("Hooked " + toolType.Name);
            
               
            }

            //elevtion limits (AI part)
            //if (elevationLimitChecksDisabled)
            //{
                Util.tryMethodRedirection(typeof(RoadAI), "GetElevationLimits", targetType, "GetElevationLimits", ref redirects);
                Util.tryMethodRedirection(typeof(TrainTrackAI), "GetElevationLimits", targetType, "GetElevationLimits", ref redirects);
                Util.tryMethodRedirection(typeof(PedestrianPathAI), "GetElevationLimits", targetType, "GetElevationLimits", ref redirects);
                Util.tryMethodRedirection(typeof(NetAI), "BuildUnderground", targetType, "BuildUnderground", ref redirects);
            //}
            //for (int i = 0; i < detouredNetAItypes.Length; ++i)
            //    Util.tryMethodRedirection(detouredNetAItypes[i], "GetElevationLimits", targetType, "GetElevationLimits", ref redirects);

            //disable building collisions
            if (disableObjectCollisions)
            {
                Util.tryMethodRedirection(typeof(BuildingTool), "CheckCollidingBuildings", targetType, "CheckCollidingBuildings", ref redirects);
                Util.tryMethodRedirection(typeof(BuildingTool), "CheckSpace", targetType, "CheckSpace", ref redirects);
                //Util.tryMethodRedirection(typeof(TreeTool), "CheckPlacementErrors", targetType, "CheckPlacementErrorsTree", ref redirects);
                //Util.tryMethodRedirection(typeof(PropTool), "CheckPlacementErrors", targetType, "CheckPlacementErrorsProp", ref redirects);         
            }
                

            anarchyEnabled = true;
            //Debug.Log("Enabled road anarchy.");
        }


        private void disableSegmentSnapping()
        {
            if (!snappingEnabled)
                return;

            //Debug.Log("Disabling segment snapping for " + netInfos.Length + " NetInfos...");

            System.Type targetType = this.GetType();
            
            //for (int i = 0; i < netInfos.Length; ++i )
            //    Util.tryMethodRedirection(netInfos.GetType(), "GetMinNodeDistance", targetType, "GetMinNodeDistance", ref redirectsSnapping);

            Util.tryMethodRedirection(typeof(NetInfo), "GetMinNodeDistance", targetType, "GetMinNodeDistance", ref redirectsSnapping);
            
            snappingEnabled = false;
        }

        public void enableSegmentSnapping()
        {
            if (snappingEnabled)
                return;
            
            foreach (var kvp in redirectsSnapping)
                Detours.RedirectionHelper.RevertRedirect(kvp.Key, kvp.Value);
            redirectsSnapping.Clear();

            snappingEnabled = true;
        }


        private static void MoveEndNode(ref ushort node, ref Vector3 direction, Vector3 position)
        {
        }

        private static void Snap(NetInfo info, ref Vector3 point, ref Vector3 direction, Vector3 refPoint, float refAngle)    
        {
            Debug.Log("SharpJunctionAngleBehaviour::Snap()");
        }

        public static NetTool.ControlPoint SnapDirection(NetTool.ControlPoint newPoint, NetTool.ControlPoint oldPoint, NetInfo info, out bool success, out float minDistanceSq)
        {
            Debug.Log("SharpJunctionAngleBehaviour::SnapDirection()");
            success = true;
            minDistanceSq = 0.0f;
            return oldPoint;
        }

        public void disableAnarchy()
        {
            Log("disable Anarchy");

            if (!anarchyEnabled)
                return;
            
            //DebugOutputPanel.AddMessage(PluginManager.MessageType.Message, "Disabling Hook");
            
            foreach (var kvp in redirects)
                Detours.RedirectionHelper.RevertRedirect(kvp.Key, kvp.Value);
            redirects.Clear();
            
            anarchyEnabled = false;
            //Debug.Log("Disabled road anarchy.");
        }


        public void disableBending()
        {
            for (int i = 0; i < bendingNetInfos.Length; ++i)
                bendingNetInfos[i].m_enableBendingSegments = false;
            
            //Debug.Log("Disabled segment bending.");
            bendingDisabled = true;
        }


        public void enableBending()
        {
            for (int i = 0; i < bendingNetInfos.Length; ++i)
                bendingNetInfos[i].m_enableBendingSegments = true;
            
            //Debug.Log("Enabled segment bending.");
            bendingDisabled = false;
        }

        public void disableGroundInteraction()
        {
            for (int i = 0; i < netInfos.Length; ++i)
            {
                netInfos[i].m_lowerTerrain      = false;
                netInfos[i].m_flattenTerrain    = false;
                //netInfos[i].m_followTerrain     = false;
                netInfos[i].m_clipTerrain     = false;
            }
        }

        public void enableGroundInteraction()
        {
            for (int i = 0; i < netInfos.Length; ++i)
            {
                netInfos[i].m_lowerTerrain   = true;
                netInfos[i].m_flattenTerrain = true;
                //netInfos[i].m_followTerrain  = true;
                netInfos[i].m_clipTerrain  = true;
            }

        }


        private static ToolBase.ToolErrors CanCreateSegment7(NetInfo segmentInfo, ushort startNode, ushort upgrading, Vector3 endPos, Vector3 startDir, Vector3 endDir, ulong[] collidingSegmentBuffer)
        {
            return ToolBase.ToolErrors.None;
        }
      
       
        private static ToolBase.ToolErrors CanCreateSegment12(NetInfo segmentInfo, ushort startNode, ushort startSegment, ushort endNode, ushort endSegment, ushort upgrading, Vector3 startPos, Vector3 endPos, Vector3 startDir, Vector3 endDir, ulong[] collidingSegmentBuffer, bool testEnds)
        {
            return ToolBase.ToolErrors.None;
        }

       

        private static ToolBase.ToolErrors CanCreateSegment13(ushort segment, ushort endNode, ushort otherNode, NetInfo info1, Vector3 startPos, Vector3 endPos, Vector3 startDir, Vector3 endDir, NetInfo info2, Vector3 endPos2, Vector3 startDir2, Vector3 endDir2, ulong[] collidingSegmentBuffer)
        {
            return ToolBase.ToolErrors.None;
        }   
        

        private static bool CheckStartAndEnd(ushort upgrading, ushort startSegment, ushort startNode, ushort endSegment, ushort endNode, ulong[] collidingSegmentBuffer)
        {
            return true;
        }
        

        //18 params
        //public ToolBase.ToolErrors CreateNode(NetInfo info, NetTool.ControlPoint startPoint, NetTool.ControlPoint middlePoint, NetTool.ControlPoint endPoint, FastList<NetTool.NodePosition> nodeBuffer, int maxSegments, bool test, bool visualize, bool autoFix, bool needMoney, bool invert, bool switchDir, ushort relocateBuildingID, out ushort firstNode, out ushort lastNode, out ushort segment, out int cost, out int productionRate)
        //{
        //    //revert the redirection to call the original method
        //    RedirectionHelper.RevertRedirect(createNodeMethodInfo, createNodeRedirectCallState);

        //    NetTool.CreateNode(info, startPoint, middlePoint, endPoint, nodeBuffer, maxSegments, test, visualize, autoFix, needMoney, invert, switchDir, relocateBuildingID, out firstNode, out lastNode, out segment, out cost, out productionRate);

        //    //reestablish the redirection
        //    var allFlags = BindingFlags.Instance | BindingFlags.Static | BindingFlags.NonPublic | BindingFlags.Public;
        //    var method = typeof(NetTool).GetMethods(allFlags).Single(c => c.Name == "CreateNode" && c.GetParameters().Length == 18);
        //    createNodeMethodInfo = method;
        //    createNodeRedirectCallState = RedirectionHelper.RedirectCalls(method, typeof(Anarchy).GetMethod("CreateNode", allFlags));
        //    redirects.Add(createNodeMethodInfo, createNodeRedirectCallState);

        //    return ToolBase.ToolErrors.None;
        //}
   
                
        public static bool CheckCollidingBuildings(BuildingInfo other, ulong[] buildingMask, ulong[] segmentMask)
        {
            //Debug.Log("SharpJunctionAngleBehaviour::CheckCollidingBuildings()");
            return false;
        }

        public static bool CheckCollidingSegments(ulong[] segmentMask, ulong[] buildingMask, ushort upgrading)
        {
            //Debug.Log("SharpJunctionAngleBehaviour::CheckCollidingSegments()");
            return false;
        }

        private ToolBase.ToolErrors CheckSpace(BuildingInfo info, int relocating, Vector3 pos, float minY, float maxY, float angle, int width, int length, bool test, ulong[] collidingSegmentBuffer, ulong[] collidingBuildingBuffer)
        {
            //Debug.Log("SharpJunctionAngleBehaviour::CheckSpace()");
            return ToolBase.ToolErrors.None;
        }

        public bool CheckZoning(ItemClass.Zone zone)
        {
            //Debug.Log("SharpJunctionAngleBehaviour::CheckZoning()");
            return true;
        }


         private static bool CanAddNode4(ushort segmentID, ushort nodeID, Vector3 position, Vector3 direction)
         {
             //Debug.Log("SharpJunctionAngleBehaviour::CanAddNode4()");
             return true;
         }

         private static bool CanAddNode5(ushort segmentID, Vector3 position, Vector3 direction, bool checkDirection, ulong[] collidingSegmentBuffer)
         {
             //Debug.Log("SharpJunctionAngleBehaviour::CanAddNode5()");
             return true;
         }

         //private static ToolBase.ToolErrors TestNodeBuilding(BuildingInfo info, Vector3 position, Vector3 direction, ushort ignoreNode, ushort ignoreSegment, ushort ignoreBuilding, bool test, ulong[] collidingSegmentBuffer, ulong[] collidingBuildingBuffer)
         //{
         //    //Debug.Log("SharpJunctionAngleBehaviour::TestNodeBuilding()");
         //    return ToolBase.ToolErrors.None;
         //}

         private static ToolBase.ToolErrors TestNodeBuilding(BuildingInfo info, Vector3 position, Vector3 direction, ushort ignoreNode, ushort ignoreSegment, ushort ignoreBuilding, bool test, ulong[] collidingSegmentBuffer, ulong[] collidingBuildingBuffer)
         {
             Vector2 vector = new Vector2(direction.x, direction.z) * ((float)info.m_cellLength * 4f - 0.8f);
             Vector2 vector2 = new Vector2(direction.z, -direction.x) * ((float)info.m_cellWidth * 4f - 0.8f);
             if (info.m_circular)
             {
                 vector2 *= 0.7f;
                 vector *= 0.7f;
             }
             Vector2 a = VectorUtils.XZ(position);
             Quad2 quad = default(Quad2);
             quad.a = a - vector2 - vector;
             quad.b = a - vector2 + vector;
             quad.c = a + vector2 + vector;
             quad.d = a + vector2 - vector;

            ItemClass.CollisionType collisionType = ItemClass.CollisionType.Terrain;
            if (info.m_class.m_layer == ItemClass.Layer.WaterPipes)
                collisionType = ItemClass.CollisionType.Underground;

             ToolBase.ToolErrors toolErrors = ToolBase.ToolErrors.None;
             float minY = Mathf.Min(position.y, Singleton<TerrainManager>.instance.SampleRawHeightSmooth(position));
             float maxY = position.y + info.m_generatedInfo.m_size.y;
             Singleton<NetManager>.instance.OverlapQuad(quad, minY, maxY, collisionType, info.m_class.m_layer, ignoreNode, 0, ignoreSegment, collidingSegmentBuffer);
             Singleton<BuildingManager>.instance.OverlapQuad(quad, minY, maxY, collisionType, info.m_class.m_layer, ignoreBuilding, ignoreNode, 0, collidingBuildingBuffer);
             if (!Singleton<BuildingManager>.instance.CheckLimits())
             {
                 toolErrors |= ToolBase.ToolErrors.TooManyObjects;
             }
             return toolErrors;
         }

         private static bool CanAddSegment(ushort nodeID, Vector3 direction, ulong[] collidingSegmentBuffer, ushort ignoreSegment)
         {
             //Debug.Log("SharpJunctionAngleBehaviour::CanAddSegment()");
             return true;
         }

         public static ToolBase.ToolErrors CheckPlacementErrorsTree(TreeInfo info, Vector3 position, bool fixedHeight, uint id, ulong[] collidingSegmentBuffer, ulong[] collidingBuildingBuffer)
         {
             return ToolBase.ToolErrors.None;
         }

         public static ToolBase.ToolErrors CheckPlacementErrorsProp(PropInfo info, Vector3 position, bool fixedHeight, ushort id, ulong[] collidingSegmentBuffer, ulong[] collidingBuildingBuffer)
         {
             return ToolBase.ToolErrors.None;
         }


        private float GetElevation(NetInfo info)
        {
            //Debug.Log("SharpJunctionAngleBehaviour::GetElevation()");
            var ele = (NetTool)ToolManager.instance.m_properties.CurrentTool;
            var mi = ele.GetType().GetField("m_elevation", BindingFlags.Instance | BindingFlags.NonPublic);
            return (float)Mathf.Clamp((int)mi.GetValue(ele), -512, 1920) * 6f;
        }
        
        public bool BuildUnderground()
        {
            //Debug.Log("SharpJunctionAngleBehaviour::BuildUnderground()");
            return true;
        }


        public void GetElevationLimits(out int min, out int max)
        {
            min = MIN_HEIGHT;
            max = MAX_HEIGHT;
            //min = Mathf.RoundToInt(-120 / terrainStep);
            //max = Mathf.RoundToInt(255 / terrainStep);
        }
        
        public static void GetAdjustedElevationLimits(NetAI netAI, out int min, out int max)
        {            
            min = MIN_HEIGHT;
            max = MAX_HEIGHT;
            //netAI.GetElevationLimits(out min, out max);
            //min *= 2;
            //max *= 2;
            //min = Mathf.RoundToInt(min * 12 / 3);
            //max = Mathf.RoundToInt(max * 12 / 3);
        }


        //this cant work, because this pointer (0th param) gets pointless after redirection
        //public void GetAdjustedElevationLimits(NetAI netAI, out int min, out int max)
        //{
        //    min = Mathf.RoundToInt(-120 / terrainStep);
        //    max = Mathf.RoundToInt(255 / terrainStep);
        //}

        public float GetMinNodeDistance()
        {
           return 3f;
        }

        private static ToolBase.ToolErrors CheckNodeHeightsSimple(NetInfo info, FastList<NetTool.NodePosition> nodeBuffer)
        {
            //Debug.Log("SharpJunctionAngleBehaviour::CheckNodeHeights()");
            return ToolBase.ToolErrors.None;
        }

        private static ToolBase.ToolErrors CheckNodeHeights(NetInfo info, FastList<NetTool.NodePosition> nodeBuffer)
        {
            bool flag = info.m_netAI.BuildUnderground();
            bool flag2 = info.m_netAI.SupportUnderground();
            if (info.m_netAI.LinearMiddleHeight())
            {
                if (nodeBuffer.m_size >= 3)
                {
                    Vector2 b = VectorUtils.XZ(nodeBuffer.m_buffer[0].m_position);
                    Vector2 b2 = VectorUtils.XZ(nodeBuffer.m_buffer[nodeBuffer.m_size - 1].m_position);
                    float y = nodeBuffer.m_buffer[0].m_position.y;
                    float y2 = nodeBuffer.m_buffer[nodeBuffer.m_size - 1].m_position.y;
                    for (int i = 1; i < nodeBuffer.m_size - 1; i++)
                    {
                        NetTool.NodePosition nodePosition = nodeBuffer.m_buffer[i];
                        float num = Vector2.Distance(VectorUtils.XZ(nodePosition.m_position), b);
                        float num2 = Vector2.Distance(VectorUtils.XZ(nodePosition.m_position), b2);
                        nodePosition.m_position.y = Mathf.Lerp(y, y2, num / Mathf.Max(1f, num + num2));
                        nodeBuffer.m_buffer[i] = nodePosition;
                    }
                }

                return ToolBase.ToolErrors.None;
            }
            else
            {
                bool flag3 = false;
                for (int j = 1; j < nodeBuffer.m_size; j++)
                {
                    NetTool.NodePosition nodePosition2 = nodeBuffer.m_buffer[j - 1];
                    NetTool.NodePosition nodePosition3 = nodeBuffer.m_buffer[j];
                    float num4 = VectorUtils.LengthXZ(nodePosition3.m_position - nodePosition2.m_position);
                    float num5 = num4 * info.m_maxSlope;
                    nodePosition3.m_minY = Mathf.Max(nodePosition3.m_minY, nodePosition2.m_minY - num5);
                    nodePosition3.m_maxY = Mathf.Min(nodePosition3.m_maxY, nodePosition2.m_maxY + num5);
                    if (!flag2)
                    {
                        nodePosition3.m_terrainHeight = Mathf.Min(nodePosition3.m_terrainHeight, nodePosition3.m_position.y + 7.98f);
                        nodePosition3.m_minY = Mathf.Max(nodePosition3.m_minY, nodePosition3.m_terrainHeight - 7.99f);
                    }
                    nodeBuffer.m_buffer[j] = nodePosition3;
                }
                for (int k = nodeBuffer.m_size - 2; k >= 0; k--)
                {
                    NetTool.NodePosition nodePosition4 = nodeBuffer.m_buffer[k + 1];
                    NetTool.NodePosition nodePosition5 = nodeBuffer.m_buffer[k];
                    float num6 = VectorUtils.LengthXZ(nodePosition5.m_position - nodePosition4.m_position);
                    float num7 = num6 * info.m_maxSlope;
                    nodePosition5.m_minY = Mathf.Max(nodePosition5.m_minY, nodePosition4.m_minY - num7);
                    nodePosition5.m_maxY = Mathf.Min(nodePosition5.m_maxY, nodePosition4.m_maxY + num7);
                    nodeBuffer.m_buffer[k] = nodePosition5;
                }
                for (int l = 0; l < nodeBuffer.m_size; l++)
                {
                    NetTool.NodePosition nodePosition6 = nodeBuffer.m_buffer[l];
                    if (nodePosition6.m_minY > nodePosition6.m_maxY)
                    {
                        return ToolBase.ToolErrors.None; /*SlopeTooSteep;*/
                    }
                    if (nodePosition6.m_position.y > nodePosition6.m_maxY)
                    {
                        nodePosition6.m_position.y = nodePosition6.m_maxY;
                        if (!flag && nodePosition6.m_elevation >= -8f)
                        {
                            nodePosition6.m_minY = nodePosition6.m_maxY;
                        }
                        flag3 = true;
                    }
                    else if (nodePosition6.m_position.y < nodePosition6.m_minY)
                    {
                        nodePosition6.m_position.y = nodePosition6.m_minY;
                        if (flag || nodePosition6.m_elevation < -8f)
                        {
                            nodePosition6.m_maxY = nodePosition6.m_minY;
                        }
                        flag3 = true;
                    }
                    nodeBuffer.m_buffer[l] = nodePosition6;
                }
                if (nodeBuffer.m_size << 1 == 0)
                {
                    return ToolBase.ToolErrors.None; /*SlopeTooSteep;*/
                }
                if (!flag3)
                {
                    for (int m = 1; m < nodeBuffer.m_size - 1; m++)
                    {
                        NetTool.NodePosition nodePosition7 = nodeBuffer.m_buffer[m - 1];
                        NetTool.NodePosition nodePosition8 = nodeBuffer.m_buffer[m];
                        float num8 = VectorUtils.LengthXZ(nodePosition8.m_position - nodePosition7.m_position);
                        float num9 = num8 * info.m_maxSlope;
                        if (flag || nodePosition8.m_elevation < -8f)
                        {
                            if (nodePosition8.m_position.y > nodePosition7.m_position.y + num9)
                            {
                                nodePosition8.m_position.y = nodePosition7.m_position.y + num9;
                            }
                        }
                        else if (nodePosition8.m_position.y < nodePosition7.m_position.y - num9)
                        {
                            nodePosition8.m_position.y = nodePosition7.m_position.y - num9;
                        }
                        nodeBuffer.m_buffer[m] = nodePosition8;
                    }
                    for (int n = nodeBuffer.m_size - 2; n > 0; n--)
                    {
                        NetTool.NodePosition nodePosition9 = nodeBuffer.m_buffer[n + 1];
                        NetTool.NodePosition nodePosition10 = nodeBuffer.m_buffer[n];
                        float num10 = VectorUtils.LengthXZ(nodePosition10.m_position - nodePosition9.m_position);
                        float num11 = num10 * info.m_maxSlope;
                        if (flag || nodePosition10.m_elevation < -8f)
                        {
                            if (nodePosition10.m_position.y > nodePosition9.m_position.y + num11)
                            {
                                nodePosition10.m_position.y = nodePosition9.m_position.y + num11;
                            }
                        }
                        else if (nodePosition10.m_position.y < nodePosition9.m_position.y - num11)
                        {
                            nodePosition10.m_position.y = nodePosition9.m_position.y - num11;
                        }
                        nodeBuffer.m_buffer[n] = nodePosition10;
                    }
                    int num12;
                    int num13;
                    info.m_netAI.GetElevationLimits(out num12, out num13);
                    if (num13 > num12 && !flag)
                    {
                        int num15;
                        for (int num14 = 0; num14 < nodeBuffer.m_size - 1; num14 = num15)
                        {
                            NetTool.NodePosition nodePosition11 = nodeBuffer.m_buffer[num14];
                            num15 = num14 + 1;
                            float num16 = 0f;
                            bool flag4 = nodeBuffer.m_buffer[num15].m_position.y >= nodeBuffer.m_buffer[num15].m_terrainHeight + 8f;
                            bool flag5 = nodeBuffer.m_buffer[num15].m_position.y <= nodeBuffer.m_buffer[num15].m_terrainHeight - 8f;
                            if (!flag2)
                            {
                                flag5 = false;
                            }
                            if (flag4 || flag5)
                            {
                                while (num15 < nodeBuffer.m_size)
                                {
                                    NetTool.NodePosition nodePosition12 = nodeBuffer.m_buffer[num15];
                                    num16 += VectorUtils.LengthXZ(nodePosition12.m_position - nodePosition11.m_position);
                                    if (flag4 && nodePosition12.m_position.y < nodePosition12.m_terrainHeight + 8f)
                                    {
                                        break;
                                    }
                                    if (flag5 && nodePosition12.m_position.y > nodePosition12.m_terrainHeight - 8f)
                                    {
                                        break;
                                    }
                                    nodePosition11 = nodePosition12;
                                    if (num15 == nodeBuffer.m_size - 1)
                                    {
                                        break;
                                    }
                                    num15++;
                                }
                            }
                            float y3 = nodeBuffer.m_buffer[num14].m_position.y;
                            float y4 = nodeBuffer.m_buffer[num15].m_position.y;
                            nodePosition11 = nodeBuffer.m_buffer[num14];
                            float num17 = 0f;
                            num16 = Mathf.Max(1f, num16);
                            for (int num18 = num14 + 1; num18 < num15; num18++)
                            {
                                NetTool.NodePosition nodePosition13 = nodeBuffer.m_buffer[num18];
                                num17 += VectorUtils.LengthXZ(nodePosition13.m_position - nodePosition11.m_position);
                                if (flag5)
                                {
                                    nodePosition13.m_position.y = Mathf.Min(nodePosition13.m_position.y, Mathf.Lerp(y3, y4, num17 / num16));
                                }
                                else
                                {
                                    nodePosition13.m_position.y = Mathf.Max(nodePosition13.m_position.y, Mathf.Lerp(y3, y4, num17 / num16));
                                }
                                nodeBuffer.m_buffer[num18] = nodePosition13;
                                nodePosition11 = nodePosition13;
                            }
                        }
                    }
                }
                ToolBase.ToolErrors toolErrors = ToolBase.ToolErrors.None;
                for (int num19 = 1; num19 < nodeBuffer.m_size - 1; num19++)
                {
                    NetTool.NodePosition nodePosition14 = nodeBuffer.m_buffer[num19 - 1];
                    NetTool.NodePosition nodePosition15 = nodeBuffer.m_buffer[num19 + 1];
                    NetTool.NodePosition nodePosition16 = nodeBuffer.m_buffer[num19];
                    /*if (flag)
                    {
                        if (nodePosition16.m_terrainHeight < nodePosition16.m_position.y)
                        {
                            toolErrors |= ToolBase.ToolErrors.SlopeTooSteep;
                        }
                    }*
                    else if (nodePosition16.m_elevation < -8f)
                    {
                        if (nodePosition16.m_terrainHeight <= nodePosition16.m_position.y + 8f)
                        {
                            toolErrors |= ToolBase.ToolErrors.SlopeTooSteep;
                        }
                    }
                    else if (!flag2 && nodePosition16.m_terrainHeight > nodePosition16.m_position.y + 8f)
                    {
                        toolErrors |= ToolBase.ToolErrors.SlopeTooSteep;
                    }*/
                    nodePosition16.m_direction.y = VectorUtils.NormalizeXZ(nodePosition15.m_position - nodePosition14.m_position).y;
                    nodeBuffer.m_buffer[num19] = nodePosition16;
                }
                return toolErrors;
            }
        }
               

    }



}